# encoding: utf-8

# -------------------------------------------------------------------------------
# (C) Copyright 2019 - 2022 Micro Focus or one of its affiliates.
#
# The only warranties for products and services of Micro Focus and its affiliates
# and licensors (“Micro Focus”) are as may be set forth in the express warranty
# statements accompanying such products and services. Nothing herein should be
# construed as constituting an additional warranty. Micro Focus shall not be
# liable for technical or editorial errors or omissions contained herein. The
# information contained herein is subject to change without notice.
#
# Except as specifically indicated otherwise, this document contains confidential
# information and a valid license is required for possession, use or copying. If
# this work is provided to the U.S. Government, consistent with FAR 12.211 and
# 12.212, Commercial Computer Software, Computer Software Documentation, and
# Technical Data for Commercial Items  are licensed to the U.S. Government under
# vendor's standard commercial license.
# ------------------------------------------------------------------------------
# Copyright (c) 2006 Novell, Inc. All Rights Reserved.
#
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of version 2 of the GNU General Public License as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, contact Novell, Inc.
#
# To contact Novell about this file by physical or electronic mail, you may find
# current contact information at www.novell.com.
# ------------------------------------------------------------------------------

# File:	modules/NovellDhcp.ycp
# Package:	Configuration of Novell DHCP Services
# Summary:	Novell Dhcp Services settings, input and output functions
# Authors:	Mark Clark <mclark@novell.com>
#
# $Id: NovellDhcp.ycp 27914 2006-02-13 14:32:08Z locilka $
#
# Representation of the configuration of novell-dhcp.
# Input and output routines.
require "yast"

module Yast
  class NovellDhcpClass < Module
    def main
      textdomain "novell-dhcp"

      Yast.import "Popup"
      Yast.import "Service"
      Yast.import "Progress"
      Yast.import "Report"
      Yast.import "Summary"
      Yast.import "Misc"
      Yast.import "NetworkInterfaces"
      Yast.import "Edirectory"
      Yast.import "Message"
      Yast.import "Mode"
      Yast.import "NovellUtils"
      Yast.import "NovellLDAP"
      Yast.import "NovellPatterns"
      Yast.import "NovellAddonInstall"
      Yast.import "Stage"
      Yast.import "Wizard"
      Yast.import "IP"
      Yast.import "Address"

      @auto_clone = false

      @schemaLogPath = "/var/opt/novell/eDirectory/log/oes_schema.log"

      @edir_service_configured = false

      @xad_packages_installed = false
      @xadMode = false

      @xadTestMode = false

      # Network interfaces
      @ifaces = {}

      # Settings Map
      @Settings = {}
      @oldSettings = {}

      # Data was modified?
      @modified = false

      @proposal_valid = false

      # Are we in reinstall mode?
      @reinstallMode = false

      @upgradeMode = false
      @upgrade_from = ""

      # Write only, used during autoinstallation.
      # Don't run services and SuSEconfig, it's all done at one place.
      @write_only = false
      @globals_configured = false

      # Abort function
      # return boolean return true if abort
      @AbortFunction = fun_ref(method(:Modified), "boolean ()")

      @scr_conf_path = path(".sysconfig.novell-dhcp")
    end

    # Abort function
    # @return [Boolean] return true if abort
    def Abort
      return @AbortFunction.call == true if @AbortFunction != nil
      false
    end

    # Data was modified?
    # @return true if modified
    def Modified
      Builtins.y2debug("modified=%1", @modified)
      @modified
    end

    def GetTreeType
      treeType = ""
      if Mode.installation && Stage.normal && @edir_service_configured
        treeType = Convert.to_string(
          SCR.Read(path(".sysconfig.edirectory.CONFIG_EDIR_TREE_TYPE"))
        ) # next try reading the edir conf file
        Builtins.y2milestone(
          "novell-dhcp.GetTreeType (post install):tree type from edir conf file:%1",
          treeType
        )
      else
        treeType = Ops.get_string(
          Edirectory.EdirSettings,
          "CONFIG_EDIR_TREE_TYPE",
          ""
        )
        Builtins.y2milestone(
          "novell-dhcp.GetTreeType:tree type from eDir map:%1",
          treeType
        )
        if treeType == nil || treeType == ""
          treeType = Convert.to_string(
            SCR.Read(path(".sysconfig.edirectory.CONFIG_EDIR_TREE_TYPE"))
          ) # next try reading the edir conf file
          Builtins.y2milestone(
            "novell-dhcp.GetTreeType:tree type from edir conf file:%1",
            treeType
          )
        end
      end
      treeType = "" if treeType == nil

      Builtins.tolower(treeType)
    end

    def ServerIsDsfwServer(ip)
      retCode = false

      cmd = Builtins.sformat(
        "/usr/bin/ldapsearch -x -H ldaps://%1:636 -s base -b \"\" | grep 2.16.840.1.113719.1.513.7.1 > /dev/null",
        ip
      )
      Builtins.y2milestone("NovellDhcp.ServerIsDsfwServer:executing %1", cmd)
      ret = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
      retCode = true if Ops.get_integer(ret, "exit", -1) == 0 # server is a dsfw server

      Builtins.y2milestone(
        "NovellDhcp.ServerIsDsfwServer:returned %1",
        Ops.get_integer(ret, "exit", -1)
      )
      retCode
    end

    def GetContextInfo
      # dhcp-getctx.pl -s 164.99.100.123 -p 636 -l 1 -a 'cn=admin,o=novell'

      contexts = []

      # DHCP wants the remote server if availble
      server_list = NovellLDAP.GetRemoteServers(nil)
      keyIP = ""
      if Ops.greater_than(Builtins.size(server_list), 0)
        keyIP = Ops.get_string(
          Ops.get(server_list, 0, {}),
          "CONFIG_LDAP_SERVER_IP",
          ""
        )
      else
        server_list = NovellLDAP.GetLocalServers(nil)
        keyIP = Ops.get_string(
          Ops.get(server_list, 0, {}),
          "CONFIG_LDAP_SERVER_IP",
          ""
        )
      end

      server_list = Ops.get_list(
        NovellLDAP.Settings,
        "CONFIG_LDAP_SERVERS_LIST",
        []
      )
      primary_server = NovellLDAP.GetAddressServer(keyIP, server_list)

      ldapIP = Ops.get_string(primary_server, "CONFIG_LDAP_SERVER_IP", "")
      remoteServerIsDsfw = ServerIsDsfwServer(ldapIP)

      ldapPort = Builtins.sformat(
        "%1",
        Ops.get_integer(primary_server, "CONFIG_LDAP_SECURE_PORT", 636)
      )
      ldapPort = "1636" if remoteServerIsDsfw

      useSSL = "1"

      adminDn = NovellUtils.Nds2Ldap(
        Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_ADMIN_CONTEXT", "")
      )
      adminPassword = Ops.get_string(
        NovellLDAP.Settings,
        "CONFIG_LDAP_ADMIN_PASSWORD",
        ""
      )
      if @xadMode
        xad_tree_admin_context = Ops.get_string(
          NovellLDAP.Settings,
          "XAD_TREE_ADMIN_CONTEXT",
          ""
        )
        Builtins.y2milestone(
          "NovellDhcp.GetContextInfo:XAD_TREE_ADMIN_CONTEXT:%1",
          xad_tree_admin_context
        )
        if xad_tree_admin_context != nil && xad_tree_admin_context != ""
          adminDn = NovellUtils.Nds2Ldap(xad_tree_admin_context)
          adminPassword = Ops.get_string(
            NovellLDAP.Settings,
            "XAD_TREE_ADMIN_PASSWORD",
            ""
          )
        end
      end
      Builtins.y2milestone("NovellDhcp.GetContextInfo:admin:%1", adminDn)

      cmd = Builtins.sformat(
        "perl /opt/novell/dhcp/bin/dhcp-getctx.pl -s %1 -p %2 -l %3 -a '%4'",
        ldapIP,
        ldapPort,
        useSSL,
        adminDn
      )

      Builtins.y2milestone("NovellDhcp.GetContextInfo:executing: %1", cmd)

      Popup.ShowFeedback(
        _("Searching tree for locator and group objects"),
        _("This may take a while..........")
      )
      rmap = Convert.to_map(
        SCR.Execute(
          path(".target.bash_output"),
          Ops.add(Ops.add(Ops.add("OES_DATA='", adminPassword), "' "), cmd)
        )
      )
      Popup.ClearFeedback

      if Ops.get_integer(rmap, "exit", -1) != 0
        Builtins.y2milestone(
          "NovellDhcp.GetContextInfo:could not get DHCP contexts. Error: %1",
          Ops.get_integer(rmap, "exit", -1)
        )
        contexts = nil
      else
        contextFile = "/opt/novell/dhcp/bin/dhcp-ctx"
        if Ops.greater_than(SCR.Read(path(".target.size"), contextFile), 0)
          contextData = Convert.to_string(
            SCR.Read(path(".target.string"), contextFile)
          )
          contexts = Builtins.splitstring(contextData, "\n")
          Builtins.y2milestone(
            "NovellDhcp.GetContextInfo:contexts: %1",
            contexts
          )
        else
          Builtins.y2milestone(
            "NovellDhcp.GetContextInfo:script call to dhcp-getctx did not return an error, but the file '/opt/novell/dhcp/bin/dhcp-ctx' is missing."
          )
          contexts = nil
        end
      end
      deep_copy(contexts)
    end
    def ReadOption(optname, replace)
      optvalue = Convert.to_string(
        SCR.Read(Ops.add(@scr_conf_path, Builtins.topath(optname)))
      )

      optvalue = "" if optvalue == nil
      optvalue = replace if optvalue == "" && replace != nil

      optvalue
    end

    def GetDefaultInterfaces
      interfaces = ""

      NetworkInterfaces.CleanCacheRead

      Builtins.foreach(NetworkInterfaces.List("")) do |iface|
        if iface != "" && !Builtins.issubstring(iface, "lo") &&
            !Builtins.issubstring(iface, "sit")
          if interfaces == ""
            if Address.Check4(NetworkInterfaces.GetValue(iface, "IPADDR"))
              interfaces = iface
            elsif Address.Check4(NetworkInterfaces.GetValue(iface, "IPADDR"))
              interfaces = Ops.add(Ops.add(interfaces, " "), iface)
            end
          end
        end
      end
      interfaces
    end

    # Read all novell-dhcp settings
    # @return true on success
    def Read
      # NovellDhcp read dialog caption
      caption = _("Initializing OES DHCP Services Configuration")

      # TODO FIXME Set the right number of stages
      steps = 4

      sl = 500
      Builtins.sleep(sl)

      # TODO FIXME Names of real stages
      # We do not set help text here, because it was set outside
      Progress.New(
        caption,
        " ",
        steps,
        [
          # Progress stage 1/1
          _("Read OES DHCP Services settings"),
          _("Checking dependencies")
        ],
        [
          # Progress step 1/1
          _("Reading OES DHCP Services settings..."),
          _("Checking dependencies"),
          # Progress finished
          _("Finished")
        ],
        ""
      )

      # read database
      return false if Abort()
      Progress.NextStage

      # First, get the default LDAP/eDir data
      progress_orig = Progress.set(false)
      NovellLDAP.Read
      Progress.set(progress_orig)

      # See if XAD is installed. If so, default the data.
      @xad_packages_installed = NovellPatterns.IsPatternInstalled(
        "novell-domain-services"
      )

      @xad_packages_installed = true if @xadTestMode

      Builtins.y2milestone(
        "NovellDhcp:xad_packages_installed:%1",
        @xad_packages_installed
      )
      if @xad_packages_installed
        if Ops.get_string(Edirectory.EdirSettings, "XAD_SERVICE_CONFIGURED", "") == "yes" ||
            Convert.to_string(
              SCR.Read(path(".sysconfig.novell.xad.SERVICE_CONFIGURED"))
            ) == "yes"
          @xadMode = true
        end
      end
      Builtins.y2milestone("NovellDhcp:xadMode:%1", @xadMode)

      @upgrade_from = NovellUtils.UpgradeIsFrom
      if (Mode.update || Mode.autoupgrade) &&  @upgrade_from != ""
        @oldSettings = NovellUtils.GetOldSysconfigMap("novell-dhcp")
        @upgradeMode = true
      end

      Builtins.y2milestone(
        "NovellDHCP:
        upgradeFrom = #{@upgrade_from}"
      )
      Builtins.y2milestone("NovellDHCP:oldSettings: %1", @oldSettings)

      # determine if Novell DHCP Services configuration already exists
      service_configured = Builtins.tolower(
        ReadOption("SERVICE_CONFIGURED", "no")
      )
      service_previously_configured = Builtins.tolower(
        ReadOption("SERVICE_PREVIOUSLY_CONFIGURED", "no")
      )

      if @upgradeMode
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_SERVER",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_LDAP_SERVER", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_METHOD",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_LDAP_METHOD", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_DEBUGFILE",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_LDAP_DEBUGFILE", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_SERVER_CONTEXT",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_SERVER_CONTEXT", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LOCATOR_CONTEXT",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_LOCATOR_CONTEXT", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_GROUP_CONTEXT",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_GROUP_CONTEXT", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_SERVER_OBJECT_NAME",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_SERVER_OBJECT_NAME", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_USE_SECURE_PORT",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_USE_SECURE_PORT", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_USE_SECURE_PORT_FOR_CONFIG",
          Ops.get_string(
            @oldSettings,
            "CONFIG_DHCPSRV_USE_SECURE_PORT_FOR_CONFIG",
            ""
          )
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_PORT",
          Builtins.tointeger(
            Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_LDAP_PORT", "")
          )
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_USER_CONTEXT",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_REFERRALS",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_LDAP_REFERRALS", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CHECK_METHOD",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_LDAP_CHECK_METHOD", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CERTIFICATE_AUTHORITY",
          Ops.get_string(
            @oldSettings,
            "CONFIG_DHCPSRV_LDAP_CERTIFICATE_AUTHORITY",
            ""
          )
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CLIENT_CERTIFICATE",
          Ops.get_string(
            @oldSettings,
            "CONFIG_DHCPSRV_LDAP_CLIENT_CERTIFICATE",
            ""
          )
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CLIENT_KEY",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_LDAP_CLIENT_KEY", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_INTERFACES",
          Ops.get_string(@oldSettings, "CONFIG_DHCPSRV_INTERFACES", "")
        )

        if NovellUtils.Nds2Ldap(
            Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", "")
          ) ==
            NovellUtils.Nds2Ldap(
              Ops.get_string(
                NovellLDAP.Settings,
                "CONFIG_LDAP_ADMIN_CONTEXT",
                ""
              )
            )
          Ops.set(
            @Settings,
            "CONFIG_DHCPSRV_LDAP_USER_PASSWORD",
            Ops.get_string(
              NovellLDAP.Settings,
              "CONFIG_LDAP_ADMIN_PASSWORD",
              ""
            )
          )
        end
      else
        # Next get Novell DHCP Services data
        servername = ReadOption("CONFIG_DHCPSRV_LDAP_SERVER", "")
        if servername == ""
          server_list = NovellLDAP.GetLocalServers(nil)
          if Ops.greater_than(Builtins.size(server_list), 0)
            servername = Ops.get_string(
              Ops.get(server_list, 0, {}),
              "CONFIG_LDAP_SERVER_IP",
              ""
            )
          else
            server_list = NovellLDAP.GetRemoteServers(nil)
            servername = Ops.get_string(
              Ops.get(server_list, 0, {}),
              "CONFIG_LDAP_SERVER_IP",
              ""
            )
          end
        end

        ldapadmincontext = Ops.get_string(
          NovellLDAP.Settings,
          "CONFIG_LDAP_ADMIN_CONTEXT",
          ""
        )
        if @xadMode
          xad_tree_admin_context = Ops.get_string(
            NovellLDAP.Settings,
            "XAD_TREE_ADMIN_CONTEXT",
            ""
          )
          Builtins.y2milestone(
            "NovellDhcp:XAD_TREE_ADMIN_CONTEXT:%1",
            xad_tree_admin_context
          )
          if xad_tree_admin_context != nil && xad_tree_admin_context != ""
            ldapadmincontext = xad_tree_admin_context
          end
        end
        Builtins.y2milestone("NovellDhcp:admin:%1", ldapadmincontext)

        info = NovellAddonInstall.GetProductInfo("edirectory")
        if Ops.get_boolean(info, "SERVICE_CONFIGURED", false) == true
          @edir_service_configured = true
        else
          @edir_service_configured = false
        end

        edirservercontext = ""
        if Mode.installation && Stage.normal && @edir_service_configured
          edirservercontext = Misc.SysconfigRead(
            path(".sysconfig.edirectory.CONFIG_EDIR_SERVER_CONTEXT"),
            ""
          )
          Builtins.y2milestone(
            "novelldhcp:edir server context from file (post install): %1",
            edirservercontext
          )
        else
          edirservercontext = Ops.get_string(
            Edirectory.EdirSettings,
            "CONFIG_EDIR_SERVER_CONTEXT",
            ""
          )
          Builtins.y2milestone(
            "novelldhcp:edir server context from map: %1",
            edirservercontext
          )
          if edirservercontext == nil || edirservercontext == ""
            edirservercontext = Misc.SysconfigRead(
              path(".sysconfig.edirectory.CONFIG_EDIR_SERVER_CONTEXT"),
              ""
            )
            Builtins.y2milestone(
              "novelldhcp:edir server context from file: %1",
              edirservercontext
            )
          end
        end
        if edirservercontext == nil || edirservercontext == ""
          edirservercontext = NovellUtils.GetContextNDS(ldapadmincontext)
        end

        Ops.set(@Settings, "CONFIG_DHCPSRV_LDAP_SERVER", servername)
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_METHOD",
          ReadOption("CONFIG_DHCPSRV_LDAP_METHOD", "static")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_DEBUGFILE",
          ReadOption(
            "CONFIG_DHCPSRV_LDAP_DEBUGFILE",
            "/var/log/dhcp-ldap-startup.log"
          )
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_SERVER_CONTEXT",
          ReadOption("CONFIG_DHCPSRV_SERVER_CONTEXT", edirservercontext)
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LOCATOR_CONTEXT",
          ReadOption("CONFIG_DHCPSRV_LOCATOR_CONTEXT", edirservercontext)
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_GROUP_CONTEXT",
          ReadOption("CONFIG_DHCPSRV_GROUP_CONTEXT", edirservercontext)
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_SERVER_OBJECT_NAME",
          ReadOption(
            "CONFIG_DHCPSRV_SERVER_OBJECT_NAME",
            Ops.add("DHCP_", NovellUtils.GetHostname)
          )
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_USE_SECURE_PORT",
          Builtins.tolower(ReadOption("CONFIG_DHCPSRV_USE_SECURE_PORT", "yes"))
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_USE_SECURE_PORT_FOR_CONFIG",
          Builtins.tolower(
            ReadOption("CONFIG_DHCPSRV_USE_SECURE_PORT_FOR_CONFIG", "yes")
          )
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_PORT",
          Builtins.tointeger(ReadOption("CONFIG_DHCPSRV_LDAP_PORT", "636"))
        )
        Ops.set(@Settings, "CONFIG_DHCPSRV_LDAP_PORT", 1636) if @xadMode
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_USER_CONTEXT",
          ReadOption("CONFIG_DHCPSRV_LDAP_USER_CONTEXT", ldapadmincontext)
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_REFERRALS",
          Builtins.tolower(ReadOption("CONFIG_DHCPSRV_LDAP_REFERRALS", "yes"))
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CHECK_METHOD",
          Builtins.tolower(
            ReadOption("CONFIG_DHCPSRV_LDAP_CHECK_METHOD", "never")
          )
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CERTIFICATE_AUTHORITY",
          ReadOption("CONFIG_DHCPSRV_LDAP_CERTIFICATE_AUTHORITY", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CLIENT_CERTIFICATE",
          ReadOption("CONFIG_DHCPSRV_LDAP_CLIENT_CERTIFICATE", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CLIENT_KEY",
          ReadOption("CONFIG_DHCPSRV_LDAP_CLIENT_KEY", "")
        )
        Ops.set(
          @Settings,
          "CONFIG_DHCPSRV_INTERFACES",
          ReadOption("CONFIG_DHCPSRV_INTERFACES", GetDefaultInterfaces())
        )

        useCommonProxy = Ops.get_string(
          NovellLDAP.Settings,
          "CONFIG_LDAP_DO_PROXY",
          ""
        )
        Builtins.y2milestone(
          "NovellDhcp.Read:useCommonProxy: %1",
          useCommonProxy
        )

        commonProxyName = ""
        if useCommonProxy == "yes"
          commonProxyName = NovellUtils.Nds2Ldap(
            Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_PROXY_CONTEXT", "")
          )
          Builtins.y2milestone(
            "NovellDhcp.Read:commonProxyName: '%1'",
            commonProxyName
          )
          commonProxyPassword = Ops.get_string(
            NovellLDAP.Settings,
            "CONFIG_LDAP_PROXY_PASSWORD",
            ""
          )
        end

        if service_configured != "yes" # new install
          if service_previously_configured != "yes"
            @reinstallMode = false
          else
            @reinstallMode = true
          end
          Ops.set(
            @Settings,
            "CONFIG_DHCPSRV_LDAP_USER_PASSWORD",
            Ops.get_string(
              NovellLDAP.Settings,
              "CONFIG_LDAP_ADMIN_PASSWORD",
              ""
            )
          )
          if Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_USER_PASSWORD", "") == nil
            Ops.set(@Settings, "CONFIG_DHCPSRV_LDAP_USER_PASSWORD", "")
          end

          if useCommonProxy == "yes"
            Ops.set(
              @Settings,
              "CONFIG_DHCPSRV_LDAP_USER_CONTEXT",
              commonProxyName
            )
            if commonProxyPassword != ""
              Ops.set(
                @Settings,
                "CONFIG_DHCPSRV_LDAP_USER_PASSWORD",
                commonProxyPassword
              )
            else
              Ops.set(
                @Settings,
                "CONFIG_DHCPSRV_LDAP_USER_PASSWORD",
                NovellUtils.RetrieveProxyCredentials("password")
              )
            end
            Ops.set(@Settings, "CONFIG_DHCP_COMMON_PROXY", commonProxyName)
            Ops.set(
              @Settings,
              "CONFIG_DHCP_COMMON_PROXY_PASSWORD",
              Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_USER_PASSWORD", "")
            )
          end # the config file exists, so use the data. Reconfigure mode.
        else
          @reinstallMode = true

          if useCommonProxy == "yes"
            if commonProxyPassword == ""
              commonProxyPassword = NovellUtils.RetrieveProxyCredentials(
                "password"
              )
            end

            if NovellUtils.Nds2Ldap(
                Ops.get_string(
                  @Settings,
                  "CONFIG_DHCPSRV_LDAP_USER_CONTEXT",
                  ""
                )
              ) ==
                NovellUtils.Nds2Ldap(commonProxyName)
              Ops.set(
                @Settings,
                "CONFIG_DHCPSRV_LDAP_USER_PASSWORD",
                commonProxyPassword
              )
            end
            Ops.set(@Settings, "CONFIG_DHCP_COMMON_PROXY", commonProxyName)
            Ops.set(
              @Settings,
              "CONFIG_DHCP_COMMON_PROXY_PASSWORD",
              commonProxyPassword
            )
          else
            Ops.set(@Settings, "CONFIG_DHCPSRV_LDAP_USER_PASSWORD", "") # We always ask for this and never store it in /etc/sysconfig
          end
        end
      end
      @globals_configured = true
      Builtins.y2debug("DHCPSRV READ Complete: %1", @Settings)

      # Error message
      if false
        Report.Error(_("Cannot read current OES DHCP Services settings."))
      end
      Builtins.sleep(sl)

      # check dependencies
      return false if Abort()
      Progress.NextStage

      return false if Abort()
      Progress.NextStage
      Progress.Title(_("Finished"))
      Builtins.sleep(sl)

      return false if Abort()
      @modified = false
      true
    end

    # insufficient_rights_error()
    #
    # Checks a string for output indicating insufficient rights.
    def insufficient_rights_error(msg)
      return false if msg == nil || msg == ""

      idx = Builtins.search(msg, "Insufficient access")
      return true if idx != nil && Ops.greater_or_equal(idx, 0)

      idx = Builtins.search(msg, "no access (-672)")
      return true if idx != nil && Ops.greater_or_equal(idx, 0)

      false
    end

    # Write all Novell DHCP Services settings
    # @return true on success
    def Write
      Wizard.DisableAbortButton
      pattern = "novell-dhcp"
      if NovellPatterns.IsPatternInstalled(pattern) == false
        Builtins.y2milestone(
          "The required pattern, %1, is not installed.",
          pattern
        )
        if Mode.normal
          Report.Error(
            Builtins.sformat(
              _(
                "Configuration aborted. Install the required %1 pattern and try again."
              ),
              pattern
            )
          )
        end

        return true
      end

      abortInstall = Builtins.tolower(
        NovellUtils.ReadOption("ABORT_OES_INSTALL")
      )
      if abortInstall == "yes"
        Builtins.y2milestone(
          "Novell DHCP:setting ABORT_OES_INSTALL is set to 'yes'."
        )
        return true
      end

      # See if XAD is installed. If so, default the data.
      @xad_packages_installed = NovellPatterns.IsPatternInstalled(
        "novell-domain-services"
      )

      @xad_packages_installed = true if @xadTestMode

      Builtins.y2milestone(
        "NovellDhcp:xad_packages_installed:%1",
        @xad_packages_installed
      )
      if @xad_packages_installed
        if Ops.get_string(Edirectory.EdirSettings, "XAD_SERVICE_CONFIGURED", "") == "yes" ||
            Convert.to_string(
              SCR.Read(path(".sysconfig.novell.xad.SERVICE_CONFIGURED"))
            ) == "yes"
          @xadMode = true
        end
      end
      Builtins.y2milestone("NovellDhcp:xadMode:%1", @xadMode)

      sl = 500
      retval = {}

      if Mode.autoinst
        commonProxyName = NovellUtils.Nds2Ldap(
          Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_PROXY_CONTEXT", "")
        )
        if commonProxyName != ""
          Ops.set(@Settings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", commonProxyName)
        else
          Ops.set(@Settings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", NovellUtils.RetrieveProxyCredentials("username"))
        end
        commonProxyPassword = Ops.get_string(NovellLDAP.Settings,"CONFIG_LDAP_PROXY_PASSWORD","")
        if commonProxyPassword != ""
          Ops.set(@Settings, "CONFIG_DHCPSRV_LDAP_USER_PASSWORD", commonProxyPassword)
        else
          Ops.set(@Settings, "CONFIG_DHCPSRV_LDAP_USER_PASSWORD", NovellUtils.RetrieveProxyCredentials("password"))
        end
      end

      if !@upgradeMode
        # Novell Dhcp Services write dialog caption
        if !Ops.get(
            NovellAddonInstall.products_to_configure,
            "novell-dhcp",
            true
          )
          Builtins.y2milestone("Novell DHCP -- configure later")
          return true
        end

        caption = _("Saving OES DHCP Services Configuration")

        steps = 1

        Builtins.sleep(sl)

        # We do not set help text here, because it was set outside
        Progress.New(
          caption,
          " ",
          steps,
          [
            # Progress stage 1/1
            _("Write OES DHCP Services settings")
          ],
          [
            # Progress step 1/1
            _("Writing OES DHCP Services settings..."),
            # Progress finished
            _("Finished")
          ],
          ""
        )

        # write settings
        return false if Abort()
        Progress.NextStage

        SCR.Write(path(".sysconfig.novell-dhcp.SERVICE_CONFIGURED"), "no")

        reinstallStr = ""
        reinstallStr = "REINSTALL='1' " if @reinstallMode

        upgradeStr = ""
        upgradeStr = "UPGRADE='1' " if @upgradeMode

        cmd = ""
        rval = 0

        # determine IP addresses
        server_list = Ops.get_list(
          NovellLDAP.Settings,
          "CONFIG_LDAP_SERVERS_LIST",
          []
        )
        primary_server = NovellLDAP.GetAddressServer(
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_SERVER", ""),
          server_list
        )
        fallback_list = NovellLDAP.GetFallbackServerList(
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_SERVER", ""),
          server_list
        )
        Builtins.y2milestone("Primary server: %1", primary_server)
        Builtins.y2milestone("Fallback server_list: %1", fallback_list)

        full_admin_context = NovellUtils.Nds2Ldap(
          Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_ADMIN_CONTEXT", "")
        )
        admin_pwd = Ops.get_string(
          NovellLDAP.Settings,
          "CONFIG_LDAP_ADMIN_PASSWORD",
          ""
        )
        if @xadMode
          xad_tree_admin_context = Ops.get_string(
            NovellLDAP.Settings,
            "XAD_TREE_ADMIN_CONTEXT",
            ""
          )
          Builtins.y2milestone(
            "NovellDhcp:XAD_TREE_ADMIN_CONTEXT:%1",
            xad_tree_admin_context
          )
          if xad_tree_admin_context != nil && xad_tree_admin_context != ""
            full_admin_context = NovellUtils.Nds2Ldap(xad_tree_admin_context)
            admin_pwd = Ops.get_string(
              NovellLDAP.Settings,
              "XAD_TREE_ADMIN_PASSWORD",
              ""
            )
          end
        end
        Builtins.y2milestone("NovellDhcp:admin:%1", full_admin_context)
        Builtins.y2milestone("NovellDhcp:pword:%1", admin_pwd) if @xadTestMode

        # Extend Novell DHCP Services schema.
        schema_file = "/opt/novell/sch/dhcpschema.sch"
        ndssch_file = "/opt/novell/eDirectory/bin/ndssch"
        admin_con_dn = NovellUtils.Ldap2Nds(full_admin_context)

        cmd = Ops.add(
          Ops.add(
            Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(ndssch_file, " -h "),
                        Ops.get_string(
                          primary_server,
                          "CONFIG_LDAP_SERVER_IP",
                          ""
                        )
                      ),
                      " -F '"
                    ),
                    @schemaLogPath
                  ),
                  "' '"
                ),
                admin_con_dn
              ),
              "'"
            ),
            " "
          ),
          schema_file
        )

        if @xadTestMode
          Builtins.y2milestone(
            "novelldhcp:Executing %1",
            Ops.add(Ops.add(Ops.add("ADM_PASSWD='", admin_pwd), "' "), cmd)
          )
        else
          Builtins.y2milestone("novelldhcp:Executing %1", cmd)
        end

        mval = Convert.to_map(
          SCR.Execute(
            path(".target.bash_output"),
            Ops.add(Ops.add(Ops.add("ADM_PASSWD='", admin_pwd), "' "), cmd)
          )
        )
        if Ops.get_integer(mval, "exit", -1) != 0
          Builtins.y2milestone(
            "novelldhcp:Failed (error=%1) to extend schema using default IP address: %2, using file %3",
            Ops.get_integer(mval, "exit", -1),
            Ops.get_string(primary_server, "CONFIG_LDAP_SERVER_IP", ""),
            schema_file
          )

          Builtins.foreach(fallback_list) do |server_info|
            Builtins.y2milestone("novelldhcp:Retrying with fallback address.")
            cmd = Ops.add(
              Ops.add(
                Ops.add(
                  Ops.add(
                    Ops.add(
                      Ops.add(
                        Ops.add(
                          Ops.add(
                            Ops.add(ndssch_file, " -h "),
                            Ops.get_string(
                              server_info,
                              "CONFIG_LDAP_SERVER_IP",
                              ""
                            )
                          ),
                          " -F '"
                        ),
                        @schemaLogPath
                      ),
                      "' '"
                    ),
                    admin_con_dn
                  ),
                  "'"
                ),
                " "
              ),
              schema_file
            )
            if @xadTestMode
              Builtins.y2milestone(
                "novelldhcp:Executing %1",
                Ops.add(Ops.add(Ops.add("ADM_PASSWD='", admin_pwd), "' "), cmd)
              )
            else
              Builtins.y2milestone("novelldhcp:Executing %1", cmd)
            end
            mval = Convert.to_map(
              SCR.Execute(
                path(".target.bash_output"),
                Ops.add(Ops.add(Ops.add("ADM_PASSWD='", admin_pwd), "' "), cmd)
              )
            )
            if Ops.get_integer(mval, "exit", -1) != 0
              Builtins.y2milestone(
                "novelldhcp:Failed (error=%1) to extend schema using fallback IP address: %2, using file %3",
                Ops.get_integer(mval, "exit", -1),
                Ops.get_string(server_info, "CONFIG_LDAP_SERVER_IP", ""),
                schema_file
              )
            else
              raise Break
            end
          end
        end
        if Ops.get_integer(mval, "exit", -1) != 0
          Builtins.y2milestone(
            "novelldhcp.write:Failed to extend schema for Novell DHCP Services. Schema may already be extended if joining a NetWare Tree."
          )
        end

        # Call the Novell DHCP Services config module
        ldap_host = Ops.get_string(primary_server, "CONFIG_LDAP_SERVER_IP", "")
        ldap_port = Ops.get_integer(primary_server, "CONFIG_LDAP_PORT", 389)
        ldap_port = 1389 if @xadMode
        ldap_user_name = NovellUtils.Nds2Ldap(
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", "")
        )
        user_password = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_USER_PASSWORD",
          ""
        )

        useCommonProxy = Ops.get_string(
          NovellLDAP.Settings,
          "CONFIG_LDAP_DO_PROXY",
          ""
        )
        Builtins.y2milestone(
          "NovellDhcp.Write:useCommonProxy: %1",
          useCommonProxy
        )
        if useCommonProxy == "yes"
          commonProxyName = NovellUtils.Nds2Ldap(
            Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_PROXY_CONTEXT", "")
          )
          commonProxyPassword = NovellUtils.RetrieveProxyCredentials("password")
          if commonProxyPassword == nil || commonProxyPassword == ""
            commonProxyPassword = Ops.get_string(
              NovellLDAP.Settings,
              "CONFIG_LDAP_PROXY_PASSWORD",
              ""
            )
          end

          if NovellUtils.Nds2Ldap(
              Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", "")
            ) == commonProxyName
            user_password = commonProxyPassword
            Ops.set(
              @Settings,
              "CONFIG_DHCPSRV_LDAP_USER_PASSWORD",
              user_password
            )
          end
        end

        ldap_method = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_METHOD",
          "static"
        )

        ldap_debugfile = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_DEBUGFILE",
          "/var/log/dhcp-ldap-startup.log"
        )
        Builtins.y2milestone(
          "NovellDhcp.Write:ldap_debugfile: %1",
          ldap_debugfile
        )
        ldap_debugfile = NovellUtils.CleanUpString(ldap_debugfile)
        Builtins.y2milestone(
          "NovellDhcp.Write:ldap_debugfile after cleanup: %1",
          ldap_debugfile
        )
        if ldap_debugfile == nil || ldap_debugfile == ""
          ldap_debugfile = "/var/log/dhcp-ldap-startup.log"
          Ops.set(@Settings, "CONFIG_DHCPSRV_LDAP_DEBUGFILE", ldap_debugfile)
          Builtins.y2milestone(
            "NovellDhcp.Write:ldap_debugfile being set to: %1",
            ldap_debugfile
          )
        end

        server_context = NovellUtils.Nds2Ldap(
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_SERVER_CONTEXT", "")
        )
        locator_context = NovellUtils.Nds2Ldap(
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LOCATOR_CONTEXT", "")
        )
        group_context = NovellUtils.Nds2Ldap(
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_GROUP_CONTEXT", "")
        )
        server_object_name = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_SERVER_OBJECT_NAME",
          ""
        )
        ldap_referrals = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_REFERRALS",
          ""
        )
        check_method = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CHECK_METHOD",
          ""
        )
        ldap_ca = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CERTIFICATE_AUTHORITY",
          ""
        )
        client_cert = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CLIENT_CERTIFICATE",
          ""
        )
        client_key = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CLIENT_KEY",
          ""
        )
        dhcp_ldap_port = Builtins.sformat(
          "%1",
          Ops.get_integer(@Settings, "CONFIG_DHCPSRV_LDAP_PORT", 636)
        )
        #    if( xadMode )
        #        dhcp_ldap_port = "1636";

        ssl = "0"
        if Ops.get_string(@Settings, "CONFIG_DHCPSRV_USE_SECURE_PORT", "") != "no"
          ssl = "1"
        end

        ssl_conf = "0"
        if Ops.get_string(
            @Settings,
            "CONFIG_DHCPSRV_USE_SECURE_PORT_FOR_CONFIG",
            ""
          ) != "no"
          ssl_conf = "1"
          ldap_port = Ops.get_integer(
            primary_server,
            "CONFIG_LDAP_SECURE_PORT",
            636
          )
          ldap_port = 1636 if @xadMode
        end

        cmd1 = Builtins.sformat(
          "/opt/novell/dhcp/bin/dhcp_config.sh -lh %1 -lp %2 -lps %3 -lu '%4' -lpw '****' -la '%5' -lw '****' -lm %6 -ld %7",
          Ops.get_string(primary_server, "CONFIG_LDAP_SERVER_IP", ""),
          ldap_port,
          dhcp_ldap_port,
          ldap_user_name,
          full_admin_context,
          ldap_method,
          ldap_debugfile
        )

        referral = "on"
        referral = "off" if ldap_referrals == "no"

        cmd2 = Builtins.sformat(
          " -sc '%1' -lc '%2' -lg '%3' -so %4 -s %5 -ss %6 -lr %7",
          server_context,
          locator_context,
          group_context,
          server_object_name,
          ssl_conf,
          ssl,
          referral
        )

        # Send only the command that have a value.
        cmd3 = ""
        if ssl == "1"
          if check_method != ""
            cmd3 = Ops.add(Ops.add(" -lccm '", check_method), "'")
          end
          if ldap_ca != ""
            cmd3 = Ops.add(Ops.add(Ops.add(cmd3, " -lca '"), ldap_ca), "'")
          end
          if client_cert != ""
            cmd3 = Ops.add(Ops.add(Ops.add(cmd3, " -lcc '"), client_cert), "'")
          end
          if client_key != ""
            cmd3 = Ops.add(Ops.add(Ops.add(cmd3, " -lck '"), client_key), "'")
          end
        end

        if @xadTestMode
          Builtins.y2milestone(
            "novelldhcp.write:Executing %1",
            Ops.add(
              Ops.add(upgradeStr, reinstallStr),
              Builtins.sformat(
                "OES_DATA='%1' OES_DATA2='%2' %3",
                admin_pwd,
                user_password,
                Ops.add(Ops.add(cmd1, cmd2), cmd3)
              )
            )
          )
        else
          Builtins.y2milestone(
            "novelldhcp.write:Executing %1 %2 %3",
            upgradeStr,
            reinstallStr,
            Ops.add(Ops.add(cmd1, cmd2), cmd3)
          )
        end
        retval = Convert.to_map(
          SCR.Execute(
            path(".target.bash_output"),
            Ops.add(
              Ops.add(upgradeStr, reinstallStr),
              Builtins.sformat(
                "OES_DATA='%1' OES_DATA2='%2' %3",
                admin_pwd,
                user_password,
                Ops.add(Ops.add(cmd1, cmd2), cmd3)
              )
            )
          )
        )
        Builtins.y2milestone(
          "novelldhcp::dhcp_config output:Map Value:\n%1",
          retval
        )

        iRet = Ops.get_integer(retval, "exit", -1)
        if iRet != 0
          Builtins.y2milestone("novelldhcp.write:Return values: %1", retval)

          errorStr = Builtins.sformat(
            _("Failed to configure OES DHCP Services. Error: %1"),
            Ops.get_integer(retval, "exit", -1)
          )
          if iRet == 19
            errorStr = Builtins.sformat(
              _(
                "Failed to configure OES DHCP Services. Error: %1\nFailure due to non-compliance with password policy"
              ),
              Ops.get_integer(retval, "exit", -1)
            )
          end

          Builtins.y2error(errorStr)
          Report.Error(errorStr)
        end

        # Update the /etc/sysconfig/dhcpd file with the user selected interfaces
        interfaces = Ops.get_string(@Settings, "CONFIG_DHCPSRV_INTERFACES", "")
        Builtins.y2milestone(
          "novelldhcp.write:setting dhcpd file with '%1'",
          interfaces
        )
        if interfaces != nil && interfaces != ""
          retVal = SCR.Write(
            path(".sysconfig.dhcpd.DHCPD_INTERFACE"),
            interfaces
          )
          Builtins.y2milestone(
            "novelldhcp.write:return value from WriteSetting is '%1'",
            retVal
          )
        end
      else
        Builtins.y2milestone("NovellDhcp.write:config for upgrade.")
        dhcp_ldap_server = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_SERVER",
          ""
        )
        port = Ops.get_integer(@Settings, "CONFIG_DHCPSRV_LDAP_PORT", -1)
        adminDn = NovellUtils.Nds2Ldap(
          Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_ADMIN_CONTEXT", "")
        )
        adminPassword = Ops.get_string(
          NovellLDAP.Settings,
          "CONFIG_LDAP_ADMIN_PASSWORD",
          ""
        )
        if @xadMode
          xad_tree_admin_context = Ops.get_string(
            NovellLDAP.Settings,
            "XAD_TREE_ADMIN_CONTEXT",
            ""
          )
          Builtins.y2milestone(
            "NovellDhcp.GetContextInfo:XAD_TREE_ADMIN_CONTEXT:%1",
            xad_tree_admin_context
          )
          if xad_tree_admin_context != nil && xad_tree_admin_context != ""
            adminDn = NovellUtils.Nds2Ldap(xad_tree_admin_context)
            adminPassword = Ops.get_string(
              NovellLDAP.Settings,
              "XAD_TREE_ADMIN_PASSWORD",
              ""
            )
          end
        end

        dhcp_server_context = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_SERVER_CONTEXT",
          ""
        )
        dhcp_server_object = Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_SERVER_OBJECT_NAME",
          ""
        )
        ssl = "0"
        if Ops.get_string(@Settings, "CONFIG_DHCPSRV_USE_SECURE_PORT", "") != "no"
          ssl = "1"
        end
        cmd = Builtins.sformat(
          "/opt/novell/dhcp/bin/dhcp-maint -es %1 %2 %3 '%4' %5 %6 %7",
          dhcp_ldap_server,
          port,
          adminDn,
          adminPassword,
          dhcp_server_context,
          dhcp_server_object,
          ssl
        )

        Builtins.y2milestone(
          "Executing:/opt/novell/dhcp/bin/dhcp-maint -es %1 %2 %3 ***** %4 %5 %6",
          dhcp_ldap_server,
          port,
          adminDn,
          dhcp_server_context,
          dhcp_server_object,
          ssl
        )

        rmap = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
        Builtins.y2milestone("rmap %1", rmap)

        #	Running upgrade scipt for dhcp
        adminDn = NovellUtils.Ldap2Nds(adminDn)

        cmd = Builtins.sformat(
          "perl /opt/novell/dhcp/bin/upgrade_dhcp.pl -a  %1 -p '%2'",
          adminDn,
          adminPassword
        )
        Builtins.y2milestone(
          "Executing: perl /opt/novell/dhcp/bin/upgrade_dhcp.pl -a  %1 -p '*****'",
          adminDn
        )
        rmap = Convert.to_map(SCR.Execute(path(".target.bash_output"), cmd))
        Builtins.y2milestone("rmap %1", rmap)

        if Mode.update || Mode.autoupgrade

          dhcp_proxyuser = NovellUtils.Nds2Ldap(Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", ""))
          proxyUser = NovellUtils.Nds2Ldap(
                 Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_PROXY_CONTEXT", ""))

          if dhcp_proxyuser == proxyUser
            Builtins.y2milestone("NovellDhcp:Setting commonproxy username and password during upgrade")
            proxyPassword = NovellUtils.RetrieveProxyCredentials("password")
            ret = NovellUtils.StoreCredsInOCSforUpgrade(proxyUser, proxyPassword,"dhcp-ldap")
            if ret != 0
              Popup.Message(
              Builtins.sformat(_("Error storing credentials in OCS for Common Proxy User: %1"),
              ret ))
              Builtins.y2milestone(
                "Novell-dns.ConfigureProxyUser:Error storing credentials in OCS for common Proxy User: %1",
                ret)
            end
          else
            Builtins.y2milestone("NovellDhcp:DHCP using service proxy, moving to common proxy")
            NovellUtils.MovetoCommonProxyDuringUpgrade("novell-dhcp")
            Ops.set(@Settings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_PROXY_CONTEXT", ""))
          end
        end
        if Mode.update || Mode.autoupgrade
          dhcp_proxyuser = NovellUtils.Nds2Ldap(Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", ""))
          proxyUser = NovellUtils.Nds2Ldap(
                 Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_PROXY_CONTEXT", ""))

          if dhcp_proxyuser != proxyUser
            Builtins.y2milestone("NovellDhcp:DHCP using service proxy, moving to common proxy")
            NovellUtils.MovetoCommonProxyDuringUpgrade("novell-dhcp")
            Ops.set(@Settings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", Ops.get_string(NovellLDAP.Settings, "CONFIG_LDAP_PROXY_CONTEXT", ""))
          end
        end
      end

      # Write the /etc/sysconfig file
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_LDAP_SERVER"),
        Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_SERVER", "")
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_LDAP_METHOD"),
        Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_METHOD", "")
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_LDAP_DEBUGFILE"),
        Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_DEBUGFILE", "")
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_SERVER_CONTEXT"),
        NovellUtils.Nds2Ldap(
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_SERVER_CONTEXT", "")
        )
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_LOCATOR_CONTEXT"),
        NovellUtils.Nds2Ldap(
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LOCATOR_CONTEXT", "")
        )
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_GROUP_CONTEXT"),
        NovellUtils.Nds2Ldap(
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_GROUP_CONTEXT", "")
        )
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_SERVER_OBJECT_NAME"),
        Ops.get_string(@Settings, "CONFIG_DHCPSRV_SERVER_OBJECT_NAME", "")
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_USE_SECURE_PORT"),
        Ops.get_string(@Settings, "CONFIG_DHCPSRV_USE_SECURE_PORT", "")
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_USE_SECURE_PORT_FOR_CONFIG"),
        Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_USE_SECURE_PORT_FOR_CONFIG",
          ""
        )
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_LDAP_PORT"),
        Builtins.sformat(
          "%1",
          Ops.get_integer(@Settings, "CONFIG_DHCPSRV_LDAP_PORT", -1)
        )
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_LDAP_USER_CONTEXT"),
        NovellUtils.Nds2Ldap(
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", "")
        )
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_LDAP_REFERRALS"),
        Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_REFERRALS", "")
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_LDAP_CHECK_METHOD"),
        Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_CHECK_METHOD", "")
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_LDAP_CERTIFICATE_AUTHORITY"),
        Ops.get_string(
          @Settings,
          "CONFIG_DHCPSRV_LDAP_CERTIFICATE_AUTHORITY",
          ""
        )
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_LDAP_CLIENT_CERTIFICATE"),
        Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_CLIENT_CERTIFICATE", "")
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_LDAP_CLIENT_KEY"),
        Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_CLIENT_KEY", "")
      )
      SCR.Write(
        path(".sysconfig.novell-dhcp.CONFIG_DHCPSRV_INTERFACES"),
        Ops.get_string(@Settings, "CONFIG_DHCPSRV_INTERFACES", "")
      )

      # open port 67
      NovellUtils.SetFirewallPort(67, :tcp, true)

      if Mode.normal
        if Service.Status("boot.apparmor") == 0
          Service.Enable("boot.apparmor")
          Service.Reload("boot.apparmor")
          Builtins.y2milestone("inst_oes_postconfig:restarted apparmor")
        else
          Builtins.y2milestone(
            "inst_oes_postconfig:apparmor not running,profiles will not be reloaded"
          )
        end
      end

      if !@upgradeMode
        if Ops.get_integer(retval, "exit", -1) == 0
          SCR.Write(path(".sysconfig.novell-dhcp.SERVICE_CONFIGURED"), "yes")
          SCR.Write(
            path(".sysconfig.novell-dhcp.SERVICE_PREVIOUSLY_CONFIGURED"),
            "yes"
          )
        end
      else
        SCR.Write(path(".sysconfig.novell-dhcp.SERVICE_CONFIGURED"), "yes")
        SCR.Write(
          path(".sysconfig.novell-dhcp.SERVICE_PREVIOUSLY_CONFIGURED"),
          "yes"
        )
      end

      # Flush the sysconfig file.
      SCR.Write(path(".sysconfig.novell-dhcp"), nil)

      Builtins.y2debug("Novell DHCP Services Settings(Write): %1", @Settings)

      return false if Abort()
      Progress.Finish
      Builtins.sleep(sl)

      return false if Abort()
      true
    end

    # Get all novell-dhcp settings from the first parameter
    # (For use by autoinstallation.)
    # @param [Hash] settings The YCP structure to be imported.
    # @return [Boolean] True on success
    def Import(settings)
      settings = deep_copy(settings)
      if settings == {}
        @globals_configured = false
        Ops.set(NovellAddonInstall.products_to_configure, "novell-dhcp", false)
        return true
      end

      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_LDAP_SERVER",
        Ops.get_string(settings, "ldap_server", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_LDAP_METHOD",
        Ops.get_string(settings, "ldap_method", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_LDAP_DEBUGFILE",
        Ops.get_string(
          settings,
          "ldap_debug_file",
          "/var/log/dhcp-ldap-startup.log"
        )
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_SERVER_CONTEXT",
        Ops.get_string(settings, "server_context", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_LOCATOR_CONTEXT",
        Ops.get_string(settings, "locator_context", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_GROUP_CONTEXT",
        Ops.get_string(settings, "group_context", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_SERVER_OBJECT_NAME",
        Ops.get_string(settings, "server_object_name", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_USE_SECURE_PORT",
        Ops.get_string(settings, "use_secure_port", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_USE_SECURE_PORT_FOR_CONFIG",
        Ops.get_string(settings, "use_secure_port_config", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_LDAP_PORT",
        Ops.get_integer(settings, "dhcp_ldap_port", -1)
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_LDAP_REFERRALS",
        Ops.get_string(settings, "ldap_referrals", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_LDAP_CHECK_METHOD",
        Ops.get_string(settings, "check_method", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_LDAP_CERTIFICATE_AUTHORITY",
        Ops.get_string(settings, "certificate_authority", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_LDAP_CLIENT_CERTIFICATE",
        Ops.get_string(settings, "client_certificate", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_LDAP_CLIENT_KEY",
        Ops.get_string(settings, "client_key", "")
      )
      Ops.set(
        @Settings,
        "CONFIG_DHCPSRV_INTERFACES",
        Ops.get_string(settings, "interfaces", "")
      )

      @globals_configured = true
      true
    end

    # Dump the novell-dhcp settings to a single map
    # (For use by autoinstallation.)
    # @return [Hash] Dumped settings (later acceptable by Import ())
    def Export
      settings = {}

      return settings if NovellUtils.CloneSettingstoAutoinstXml("novell-dhcp") == false

      if @auto_clone
        Ops.set(NovellAddonInstall.products_to_configure, "novell-dhcp", true)
      end
      if Ops.get(NovellAddonInstall.products_to_configure, "novell-dhcp", true)
        Ops.set(
          settings,
          "ldap_server",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_SERVER", "")
        )
        Ops.set(
          settings,
          "ldap_method",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_METHOD", "")
        )
        Ops.set(
          settings,
          "ldap_debug_file",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_DEBUGFILE", "")
        )
        Ops.set(
          settings,
          "server_context",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_SERVER_CONTEXT", "")
        )
        Ops.set(
          settings,
          "locator_context",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LOCATOR_CONTEXT", "")
        )
        Ops.set(
          settings,
          "group_context",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_GROUP_CONTEXT", "")
        )
        Ops.set(
          settings,
          "server_object_name",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_SERVER_OBJECT_NAME", "")
        )
        Ops.set(
          settings,
          "use_secure_port",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_USE_SECURE_PORT", "")
        )
        Ops.set(
          settings,
          "use_secure_port_config",
          Ops.get_string(
            @Settings,
            "CONFIG_DHCPSRV_USE_SECURE_PORT_FOR_CONFIG",
            ""
          )
        )
        Ops.set(
          settings,
          "dhcp_ldap_port",
          Ops.get_integer(@Settings, "CONFIG_DHCPSRV_LDAP_PORT", -1)
        )
        Ops.set(
          settings,
          "ldap_referrals",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_REFERRALS", "")
        )
        Ops.set(
          settings,
          "check_method",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_CHECK_METHOD", "")
        )
        Ops.set(
          settings,
          "certificate_authority",
          Ops.get_string(
            @Settings,
            "CONFIG_DHCPSRV_LDAP_CERTIFICATE_AUTHORITY",
            ""
          )
        )
        Ops.set(
          settings,
          "client_certificate",
          Ops.get_string(
            @Settings,
            "CONFIG_DHCPSRV_LDAP_CLIENT_CERTIFICATE",
            ""
          )
        )
        Ops.set(
          settings,
          "client_key",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_CLIENT_KEY", "")
        )
        Ops.set(
          settings,
          "interfaces",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_INTERFACES", "")
        )
      end

      deep_copy(settings)
    end

    # Dump all of the novell-dhcp settings to a single map
    # (For use by autoinstallation.)
    # @return [Hash] Dumped settings (later acceptable by Import ())
    def ExportSettings
      settings = {}

      return settings if NovellUtils.CloneSettingstoAutoinstXml("novell-dhcp") == false

      if @auto_clone
        Ops.set(NovellAddonInstall.products_to_configure, "novell-dhcp", true)
      end
      if Ops.get(NovellAddonInstall.products_to_configure, "novell-dhcp", true)
        Ops.set(
          settings,
          "ldap_server",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_SERVER", "")
        )
        Ops.set(
          settings,
          "ldap_method",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_METHOD", "")
        )
        Ops.set(
          settings,
          "ldap_debug_file",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_DEBUGFILE", "")
        )
        Ops.set(
          settings,
          "server_context",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_SERVER_CONTEXT", "")
        )
        Ops.set(
          settings,
          "locator_context",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LOCATOR_CONTEXT", "")
        )
        Ops.set(
          settings,
          "group_context",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_GROUP_CONTEXT", "")
        )
        Ops.set(
          settings,
          "server_object_name",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_SERVER_OBJECT_NAME", "")
        )
        Ops.set(
          settings,
          "use_secure_port",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_USE_SECURE_PORT", "")
        )
        Ops.set(
          settings,
          "use_secure_port_config",
          Ops.get_string(
            @Settings,
            "CONFIG_DHCPSRV_USE_SECURE_PORT_FOR_CONFIG",
            ""
          )
        )
        Ops.set(
          settings,
          "dhcp_ldap_port",
          Ops.get_integer(@Settings, "CONFIG_DHCPSRV_LDAP_PORT", -1)
        )
        Ops.set(
          settings,
          "ldap_referrals",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_REFERRALS", "")
        )
        Ops.set(
          settings,
          "check_method",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_CHECK_METHOD", "")
        )
        Ops.set(
          settings,
          "certificate_authority",
          Ops.get_string(
            @Settings,
            "CONFIG_DHCPSRV_LDAP_CERTIFICATE_AUTHORITY",
            ""
          )
        )
        Ops.set(
          settings,
          "client_certificate",
          Ops.get_string(
            @Settings,
            "CONFIG_DHCPSRV_LDAP_CLIENT_CERTIFICATE",
            ""
          )
        )
        Ops.set(
          settings,
          "client_key",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_CLIENT_KEY", "")
        )
        Ops.set(
          settings,
          "interfaces",
          Ops.get_string(@Settings, "CONFIG_DHCPSRV_INTERFACES", "")
        )
      end

      deep_copy(settings)
    end

    # Tranlate a request certificate string
    # @return translated "yes" or "no" string
    def Trans_req_cert(reqcert)
      tmp = Builtins.tolower(reqcert)
      if tmp == "allow"
        return _("allow")
      elsif tmp == "try"
        return _("try")
      elsif tmp == "hard"
        return _("hard")
      else
        return _("never")
      end
    end

    # Tranlate a yes or no string
    # @return translated "yes" or "no" string
    def Trans_yes_no(yesno)
      if Builtins.tolower(yesno) == "yes"
        return _("yes")
      else
        return _("no")
      end
    end

    # Create a textual summary and a list of unconfigured cards
    # @return summary of the current configuration
    def Summary
      summary = ""

      if !@globals_configured
        nc = Summary.NotConfigured
        summary = Summary.AddLine(summary, nc)
      else
        summary = Summary.OpenList(summary)

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("DHCP Server context: %1"),
            Ops.get_string(@Settings, "CONFIG_DHCPSRV_SERVER_CONTEXT", "")
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("DHCP Server object name: %1"),
            Ops.get_string(@Settings, "CONFIG_DHCPSRV_SERVER_OBJECT_NAME", "")
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("Locator context: %1"),
            Ops.get_string(@Settings, "CONFIG_DHCPSRV_LOCATOR_CONTEXT", "")
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("Group context: %1"),
            Ops.get_string(@Settings, "CONFIG_DHCPSRV_GROUP_CONTEXT", "")
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("Log file location: %1"),
            Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_DEBUGFILE", "")
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("LDAP method: %1"),
            Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_METHOD", "")
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("Chase LDAP referrals: %1"),
            Trans_yes_no(
              Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_REFERRALS", "")
            )
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("LDAP Server: %1"),
            Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_SERVER", "")
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("LDAP user name: %1"),
            Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_USER_CONTEXT", "")
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("Use secure channel: %1"),
            Trans_yes_no(
              Ops.get_string(@Settings, "CONFIG_DHCPSRV_USE_SECURE_PORT", "")
            )
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("Use secure channel for configuration: %1"),
            Trans_yes_no(
              Ops.get_string(
                @Settings,
                "CONFIG_DHCPSRV_USE_SECURE_PORT_FOR_CONFIG",
                ""
              )
            )
          )
        )

        summary = Summary.AddListItem(
          summary,
          Builtins.sformat(
            _("LDAP port for DHCP server: %1"),
            Ops.get_integer(@Settings, "CONFIG_DHCPSRV_LDAP_PORT", -1)
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("Request Certificate: %1"),
            Trans_req_cert(
              Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_CHECK_METHOD", "")
            )
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("LDAP CA file: %1"),
            Ops.get_string(
              @Settings,
              "CONFIG_DHCPSRV_LDAP_CERTIFICATE_AUTHORITY",
              ""
            )
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("LDAP client certificate: %1"),
            Ops.get_string(
              @Settings,
              "CONFIG_DHCPSRV_LDAP_CLIENT_CERTIFICATE",
              ""
            )
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("LDAP client Key file: %1"),
            Ops.get_string(@Settings, "CONFIG_DHCPSRV_LDAP_CLIENT_KEY", "")
          )
        )

        summary = Summary.AddListItem(
          summary, # Configuration summary text for autoyast
          Builtins.sformat(
            _("DHCP Interfaces: %1"),
            Ops.get_string(@Settings, "CONFIG_DHCPSRV_INTERFACES", "")
          )
        )

        summary = Summary.CloseList(summary)
      end
      [summary, []]
    end

    # Create an overview table with all configured cards
    # @return table items
    def Overview
      # TODO FIXME: your code here...
      []
    end

    # Return packages needed to be installed and removed during
    # Autoinstallation to insure module has all needed software
    # installed.
    # @return [Hash] with 2 lists.
    def AutoPackages
      # TODO FIXME: your code here...
      { "install" => [], "remove" => [] }
    end

    publish :variable => :auto_clone, :type => "boolean"
    publish :variable => :schemaLogPath, :type => "string"
    publish :variable => :edir_service_configured, :type => "boolean"
    publish :variable => :xad_packages_installed, :type => "boolean"
    publish :variable => :xadMode, :type => "boolean"
    publish :variable => :xadTestMode, :type => "boolean"
    publish :variable => :ifaces, :type => "map <string, map <string, any>>"
    publish :variable => :Settings, :type => "map"
    publish :variable => :oldSettings, :type => "map"
    publish :function => :Modified, :type => "boolean ()"
    publish :variable => :modified, :type => "boolean"
    publish :variable => :proposal_valid, :type => "boolean"
    publish :variable => :reinstallMode, :type => "boolean"
    publish :variable => :upgradeMode, :type => "boolean"
    publish :variable => :write_only, :type => "boolean"
    publish :variable => :globals_configured, :type => "boolean"
    publish :variable => :AbortFunction, :type => "boolean ()"
    publish :function => :Abort, :type => "boolean ()"
    publish :function => :GetTreeType, :type => "string ()"
    publish :function => :ServerIsDsfwServer, :type => "boolean (string)"
    publish :function => :GetContextInfo, :type => "list <string> ()"
    publish :function => :ReadOption, :type => "string (string, string)"
    publish :function => :Read, :type => "boolean ()"
    publish :function => :Write, :type => "boolean ()"
    publish :function => :Import, :type => "boolean (map)"
    publish :function => :Export, :type => "map ()"
    publish :function => :ExportSettings, :type => "map ()"
    publish :function => :Summary, :type => "list ()"
    publish :function => :Overview, :type => "list ()"
    publish :function => :AutoPackages, :type => "map ()"
  end

  NovellDhcp = NovellDhcpClass.new
  NovellDhcp.main
end
